/**
 * Copyright (c) 2025 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "client.h"
#include "console/console.h"
#include "assert/assert.h"

client_t CLIENT;

void clientInit(const clientinit_t init) {
  memset(&CLIENT, 0, sizeof(client_t));
  CLIENT.state = CLIENT_STATE_CONNECTING;

  CLIENT.socket = socket(AF_INET, SOCK_STREAM, 0);
  CLIENT.serverAddress.sin_family = AF_INET;
  CLIENT.serverAddress.sin_port = htons(init.port);

  if(inet_pton(AF_INET, init.host, &CLIENT.serverAddress.sin_addr) <= 0) {
    assertUnreachable("Invalid address or address not supported.");
  }

  if(connect(CLIENT.socket, (struct sockaddr *)&CLIENT.serverAddress, sizeof(CLIENT.serverAddress)) < 0) {
    assertUnreachable("Connection failed.");
  }

  pthread_create(&CLIENT.thread, NULL, clientThread, NULL);

  // Wait for client to connect
  while(CLIENT.state == CLIENT_STATE_CONNECTING) usleep(1000);

  consolePrint("Connected to server.");
}

void * clientThread(void* arg) {
  assertNull(arg, "Client thread does not accept arguments.");


  // Server sends 16 bytes.
  {
    char_t buffer[16];
    consolePrint("Reading first packet.");
    if(_clientReceive((uint8_t*)buffer, sizeof(buffer)) != sizeof(buffer)) {
      assertUnreachable("Failed to read initial server data.");
    }

    // First 4 bytes must be "Dusk"
    if(memcmp(buffer, "Dusk", 4) != 0) {
      assertUnreachable("Invalid server data.");
    }

    // Next, server will send its version number, we compare to DUSK_VERSION
    char_t cmp[] = DUSK_VERSION;
    if(memcmp(buffer+4, cmp, 4) != 0) {
      assertUnreachable("Server version mismatch.");
    }
  }

  // Send "OK" and nothing else (don't send NULL term)
  if(_clientSend((uint8_t*)"OK", 2) != 2) {
    assertUnreachable("Failed to send OK to server.");
  }

  // Server will send the first properly formed packet now.

  CLIENT.state = CLIENT_STATE_CONNECTED;
  while(CLIENT.state == CLIENT_STATE_CONNECTED) {
    usleep(1000);
  }

  CLIENT.state = CLIENT_STATE_DISCONNECTED;
  close(CLIENT.socket);
  return NULL;
}

ssize_t _clientSend(const uint8_t *data, const size_t size) {
  assertNotNull(data, "Data cannot be NULL.");
  assertTrue(size > 0, "Size cannot be zero.");

  return send(CLIENT.socket, data, size, 0);
}

ssize_t _clientReceive(uint8_t *data, const size_t size) {
  assertNotNull(data, "Data cannot be NULL.");
  assertTrue(size > 0, "Size cannot be zero.");

  fd_set readfds;
  struct timeval timeout;
  int32_t activity;

  FD_ZERO(&readfds);
  FD_SET(CLIENT.socket, &readfds);
  timeout.tv_sec = 5; // 5 seconds timeout
  timeout.tv_usec = 0;

  activity = select(CLIENT.socket + 1, &readfds, NULL, NULL, &timeout);
  if(activity <= 0) {
    assertUnreachable("Timeout or error while waiting for server data.");
  }

  return recv(CLIENT.socket, data, size, 0);
}

void clientDispose() {
  if(CLIENT.state == CLIENT_STATE_CONNECTED) {
    CLIENT.state = CLIENT_STATE_DISCONNECTING;

    // Wait for client to disconnect
    while(CLIENT.state == CLIENT_STATE_DISCONNECTING) usleep(1000);
    pthread_join(CLIENT.thread, NULL);

    consolePrint("Disconnected from server.");
  }

  close(CLIENT.socket);
}