/**
 * Copyright (c) 2025 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "serverclient.h"
#include "assert/assert.h"
#include "console/console.h"

void serverClientInit(serverclient_t *client, const serverclientinit_t init) {
  assertNotNull(client, "Client cannot be NULL.");
  client->socket = init.socket;
  consolePrint("Accepted client %i", client->socket);

  client->state = SERVER_CLIENT_STATE_CONNECTING;

  // Set socket to non-blocking
  int flags = fcntl(client->socket, F_GETFL, 0);
  fcntl(client->socket, F_SETFL, flags | O_NONBLOCK);

  // Start thread
  pthread_create(&client->thread, NULL, serverClientThread, client);

  // Wait for client to connect
  while(client->state == SERVER_CLIENT_STATE_CONNECTING) usleep(1000);
}

void * serverClientThread(void *arg) {
  serverclient_t *client = (serverclient_t*)arg;
  assertNotNull(client, "Client cannot be NULL.");

  // Send hello to client, as well as the version number.
  {
    uint8_t hello[16];
    memset(hello, 0, sizeof(hello));
    memcpy(hello, "Dusk", 4);
    strcpy((char*)&hello[4], DUSK_VERSION);
    ssize_t bytesSent = serverClientSend(client, hello, sizeof(hello));
    assertTrue(
      bytesSent == sizeof(hello),
      "Failed to send hello to client."
    );
  }

  // Client should respond "OK"
  {
    uint8_t response[2];
    ssize_t bytesReceived = serverClientReceive(client, response, sizeof(response));
    assertTrue(
      bytesReceived == sizeof(response) &&
      response[0] == 'O' &&
      response[1] == 'K',
      "Client did not respond with OK."
    );
  }

  // At this point we can start communicating "properly", so let's send the 
  // first real packet.
  packet_t packet = { .type = PACKET_TYPE_CONNECTED };
  if(!serverClientSendPacket(client, &packet)) {
    assertUnreachable("Failed to send connected packet to client.");
  }

  // Set the client state to connected
  client->state = SERVER_CLIENT_STATE_CONNECTED;
  
  while(client->state == SERVER_CLIENT_STATE_CONNECTED) {
    // Do nothing for now.
    usleep(1000);
  }

  // Disconnect
  client->state = SERVER_CLIENT_STATE_DISCONNECTED;
  return NULL;
}

ssize_t serverClientSend(
  serverclient_t *client,
  const uint8_t *data,
  const size_t size
) {
  assertNotNull(client, "Client cannot be NULL.");
  assertNotNull(data, "Data cannot be NULL.");
  assertTrue(size > 0, "Size cannot be zero.");

  return send(client->socket, data, size, 0);
}

ssize_t serverClientReceive(
  serverclient_t *client,
  uint8_t *data,
  const size_t size
) {
  assertNotNull(client, "Client cannot be NULL.");
  assertNotNull(data, "Data cannot be NULL.");
  assertTrue(size > 0, "Size cannot be zero.");

  fd_set readfds;
  struct timeval timeout;
  int32_t activity;

  FD_ZERO(&readfds);
  FD_SET(client->socket, &readfds);
  timeout.tv_sec = 5; // 5 seconds timeout
  timeout.tv_usec = 0;

  activity = select(client->socket + 1, &readfds, NULL, NULL, &timeout);
  if(activity <= 0) {
    assertUnreachable("Timeout or error while waiting for client data.");
  }

  return recv(client->socket, data, size, 0);
}

bool_t serverClientSendPacket(
  serverclient_t *client,
  const packet_t *packet
) {
  assertNotNull(client, "Client cannot be NULL.");
  assertNotNull(packet, "Packet cannot be NULL.");

  ssize_t sent = serverClientSend(client, (const uint8_t*)packet, sizeof(packet_t));
  return sent == sizeof(packet_t);
}

void serverClientDispose(serverclient_t *client) {
  assertNotNull(client, "Client cannot be NULL.");

  if(client->state == SERVER_CLIENT_STATE_CONNECTED) {
    // Let the thread know we are disconnecting
    client->state = SERVER_CLIENT_STATE_DISCONNECTING;

    // Wait for the thread to finish
    while(client->state == SERVER_CLIENT_STATE_DISCONNECTING) usleep(1000);
    pthread_join(client->thread, NULL);
  }
}